from __future__ import division, print_function
import numpy as np


def iterable(obj):
    """Utility to check if object is iterable"""
    try:
        iter(obj)
    except:
        return False
    else:
        return True


def validate_inputs(*arrays, **kwargs):
    """Validate input arrays

    This checks that
    - Arrays are mutually broadcastable
    - Broadcasted arrays are one-dimensional

    Optionally, arrays are sorted according to the ``sort_by`` argument.

    Parameters
    ----------
    *args : ndarrays
        All non-keyword arguments are arrays which will be validated
    sort_by : array
        If specified, sort all inputs by the order given in this array.
    """
    arrays = np.broadcast_arrays(*arrays)
    sort_by = kwargs.pop('sort_by', None)

    if kwargs:
        raise ValueError("unrecognized arguments: {0}".format(kwargs.keys()))

    if arrays[0].ndim != 1:
        raise ValueError("Input arrays should be one-dimensional.")

    if sort_by is not None:
        isort = np.argsort(sort_by)
        if isort.shape != arrays[0].shape:
            raise ValueError("sort shape must equal array shape.")
        arrays = tuple([a[isort] for a in arrays])
    return arrays


def windowed_sum(*arrays, **kwargs):
    """Compute the windowed sum of the given arrays

    Parameters
    ----------
    *args : array_like
        Each additional argument is an array that will be windowed
    span : int or array of ints
        The span to use for the sum at each point
    subtract_mid : boolean
        If true, then subtract the middle value from each sum
    slow : boolean
        If true, use a slow method to compute the result. This is
        primarily useful for testing and validation.

    Returns
    -------
    arrays : tuple of ndarrays
        arrays containing the windowed sum of each input array
    """
    span = kwargs.pop('span', None)
    subtract_mid = kwargs.pop('subtract_mid', False)
    slow = kwargs.pop('slow', False)

    if kwargs:
        raise ValueError("Unrecognized keywords: {0}".format(kwargs.keys()))

    if span is None:
        raise ValueError("Must provide a positive integer span")

    span = np.asarray(span, dtype=int)
    if not np.all(span > 0):
        raise ValueError("span values must be positive")

    if slow:
        # Slow fixed/variable span
        results = []
        for array in map(np.asarray, arrays):
            assert array.ndim == 1
            span, array = np.broadcast_arrays(span, array)
            result = (array[max(0, i - s // 2): i - s // 2 + s].sum()
                      for i, s, in enumerate(span))
            results.append(np.fromiter(result,
                                       dtype=array.dtype,
                                       count=len(array)))

    elif span.ndim == 0:
        # Fast fixed-span
        window = np.ones(span)
        results = [np.convolve(a, window, mode='same') for a in arrays]

    else:
        # Fast variable-span
        window = np.asarray(span)
        mins = np.arange(len(window)) - window // 2
        results = []
        for a in map(np.asarray, arrays):
            ranges = np.vstack([np.maximum(0, mins),
                                np.minimum(len(a), mins + window)]).ravel('F')
            results.append(np.add.reduceat(np.append(a, 0), ranges)[::2])

    if subtract_mid:
        results = (r - a for r, a in zip(results, arrays))

    return tuple(results)


def windowed_sum_varspan(*arrays, **kwargs):
    """TODO: document this
    TODO: combine with windowed_sum??
    """
    span = kwargs.pop('span', None)
    indices = kwargs.pop('indices', None)
    slow = kwargs.pop('slow', False)

    if kwargs:
        raise ValueError("Unrecognized keywords: {0}".format(kwargs.keys()))

    if span is None:
        raise ValueError("Must provide a positive spans")

    if indices is None:
        raise ValueError("Must provide an array of indices")

    span = np.asarray(span, dtype=int)
    if not np.all(span > 0):
        raise ValueError("span values must be positive")

    span, indices = np.broadcast_arrays(span, indices)

    if slow:
        raise ValueError("slow version not implemented")
    else:
        window = np.asarray(span)
        mins = indices - window // 2

        results = []
        for a in map(np.asarray, arrays):
            ranges = np.vstack([np.maximum(0, mins),
                                np.minimum(len(a), mins + window)]).ravel('F')
            results.append(np.add.reduceat(np.append(a, 0), ranges)[::2])

    return results


def moving_average_smooth(t, y, dy, span, cv=True, t_out=None):
    """Perform a moving-average smooth of the data

    Parameters
    ----------
    t, y, dy : array_like
        time, value, and error in value of the input data
    span : int
        the integer span of the data
    cv : boolean (default=True)
        if True, treat the problem as a cross-validation, i.e. don't use
        each point in the evaluation of its own smoothing.

    Returns
    -------
    y_smooth : array_like
        smoothed y values at each time t
    """
    t, y, dy = validate_inputs(t, y, dy, sort_by=t)
    w = dy ** -2
    w, yw = windowed_sum(w, y * w, span=span, subtract_mid=cv)

    if t_out is None:
        return yw / w
    else:
        i = np.minimum(len(t) - 1, np.searchsorted(t, t_out))
        return yw[i] / w[i]


def moving_average_smooth_varspan(t, y, dy, span, t_out):
    """
    TODO: doc
    TODO: combine with standard linear smooth?
    span matches t_out
    """
    t, y, dy = validate_inputs(t, y, dy, sort_by=t)
    indices = np.searchsorted(t, t_out)

    w = dy ** -2
    w, yw = windowed_sum_varspan(w, y * w, span=span, indices=indices)

    return yw / w


def linear_smooth(t, y, dy, span, t_out=None, cv=True):
    """Perform a linear smooth of the data

    Parameters
    ----------
    t, y, dy : array_like
        time, value, and error in value of the input data
    span : int
        the integer span of the data
    cv : boolean (default=True)
        if True, treat the problem as a cross-validation, i.e. don't use
        each point in the evaluation of its own smoothing.

    Returns
    -------
    y_smooth : array_like
        smoothed y values at each time t
    """
    t_input = t

    t, y, dy = validate_inputs(t, y, dy, sort_by=t)
    w = dy ** -2
    w, tw, yw, ttw, tyw = windowed_sum(w, t * w, y * w, t * t * w, t * y * w,
                                       span=span, subtract_mid=cv)
    denominator = (w * ttw - tw * tw)
    slope = (tyw * w - tw * yw)
    intercept = (ttw * yw - tyw * tw)

    if t_out is None:
        return (slope * t_input + intercept) / denominator
    else:
        i = np.minimum(len(t) - 1, np.searchsorted(t, t_out))
        return (slope[i] * t_out + intercept[i]) / denominator[i]


def linear_smooth_varspan(t, y, dy, span, t_out):
    """
    TODO: doc
    TODO: combine with standard linear smooth?
    span matches t_out
    """
    t, y, dy = validate_inputs(t, y, dy, sort_by=t)
    indices = np.searchsorted(t, t_out)

    w = dy ** -2
    w, tw, yw, ttw, tyw = windowed_sum_varspan(w, t * w, y * w,
                                               t * t * w, t * y * w,
                                               span=span, indices=indices)
    denominator = (w * ttw - tw * tw)
    slope = (tyw * w - tw * yw)
    intercept = (ttw * yw - tyw * tw)

    return (slope * t_out + intercept) / denominator



import numpy as np
from numpy.testing import assert_allclose, assert_array_less, assert_equal

from .. import MovingAverageSmoother, LinearSmoother


def make_sine(N=100, err=0.05, rseed=None):
    rng = np.random.RandomState(rseed)
    t = 2 * np.pi * rng.rand(N)
    y = np.sin(t) + err * rng.randn(N)
    return t, y, err


def test_variable_sine_predict():
    t, y, dy = make_sine(N=100, err=0.05, rseed=0)

    tfit = np.linspace(1, 5.3, 50)
    ytrue = np.sin(tfit)

    def check_model(Model, span):
        model1 = Model(span * np.ones_like(t))
        model2 = Model(span)
        assert_allclose(model1.fit(t, y, dy).predict(tfit),
                        model2.fit(t, y, dy).predict(tfit))

    for Model in [MovingAverageSmoother, LinearSmoother]:
        for span in [0.05, 0.2, 0.5]:
            yield check_model, Model, span


def test_variable_sine_crossval():
    t, y, dy = make_sine(N=100, err=0.05, rseed=0)

    tfit = np.linspace(1, 5.3, 50)
    ytrue = np.sin(tfit)

    def check_model(Model, span):
        model1 = Model(span * np.ones_like(t))
        model2 = Model(span)
        assert_allclose(model1.fit(t, y, dy).cv_values(),
                        model2.fit(t, y, dy).cv_values())

    for Model in [MovingAverageSmoother, LinearSmoother]:
        for span in [0.05, 0.2, 0.5]:
            yield check_model, Model, span


def test_func_span_const():
    t, y, dy = make_sine(N=100, err=0.05, rseed=0)

    def check_model(Model, span, spanfunc):
        model1 = Model(span)
        model2 = Model(spanfunc)
        assert_allclose(model1.fit(t, y, dy).predict(t),
                        model2.fit(t, y, dy).predict(t))

    for Model in [MovingAverageSmoother, LinearSmoother]:
        for span in [0.05, 0.2, 0.5]:
            spanfunc = lambda t, span=span: span * np.ones_like(t)
            yield check_model, Model, span, spanfunc


def test_func_span_variable():
    t, y, dy = make_sine(N=100, err=0.05, rseed=0)

    spanfunc = lambda t, tmax=t.max(): 0.5 + 0.05 * t / tmax
    span = spanfunc(t)

    def check_model(Model, span, spanfunc):
        model1 = Model(span)
        model2 = Model(spanfunc)
        assert_allclose(model1.fit(t, y, dy).predict(t),
                        model2.fit(t, y, dy).predict(t))

    for Model in [MovingAverageSmoother, LinearSmoother]:
        yield check_model, Model, span, spanfunc



import numpy as np
import matplotlib.pyplot as plt
from supersmoother import MovingAverageSmoother, FixedSpanSmoother


def make_test_set(N=200, err_factor=0.1, rseed=None):
    rng = np.random.RandomState(rseed)
    x = 6 * rng.rand(N)
    dy = err_factor * np.sqrt(x)
    y = np.sin(x ** 2) + dy * rng.rand(N)
    return x, y, dy


def plot_simple():
    x, y, dy = make_test_set(rseed=0)

    xfit = np.linspace(0, 6, 1000)
    model = MovingAverageSmoother(0.1).fit(x, y, dy)
    yfit = model.predict(xfit)
    yfit2 = model.predict_slow(xfit)

    print(np.allclose(yfit, yfit2))

    plt.errorbar(x, y, dy, fmt='.', alpha=0.3)
    plt.plot(xfit, yfit, '-k')
    plt.plot(xfit, yfit2, '-k')


def plot_linear():
    x, y, dy = make_test_set(rseed=0)

    xfit = np.linspace(0, 6, 1000)
    model = FixedSpanSmoother(0.05).fit(x, y, dy)
    yfit = model.predict(xfit)
    yfit2 = model.predict_slow(xfit)

    print(np.allclose(yfit, yfit2))

    plt.errorbar(x, y, dy, fmt='.', alpha=0.3)
    plt.plot(xfit, yfit, '-k')
    plt.plot(xfit, yfit2, '-k')


plt.figure()
plot_simple()

plt.figure()
plot_linear()
plt.show()



import numpy as np
from numpy.testing import assert_allclose, assert_array_less, assert_equal

from .. import MovingAverageSmoother, LinearSmoother


def make_sine(N=100, err=0.05, rseed=None):
    rng = np.random.RandomState(rseed)
    t = 2 * np.pi * rng.rand(N)
    y = np.sin(t) + err * rng.randn(N)
    return t, y, err


def make_linear(N=100, err=1E-6, rseed=None):
    rng = np.random.RandomState(rseed)
    t = 10 * rng.rand(N)
    y = t + err * rng.randn(N)
    return t, y, err


def test_sine():
    t, y, dy = make_sine(N=100, err=0.05, rseed=0)

    tfit = np.linspace(1, 5.3, 50)
    ytrue = np.sin(tfit)

    def check_model(Model, span, err):
        model = Model(span).fit(t, y, dy)
        yfit = model.predict(tfit)
        obs_err = np.mean((yfit - ytrue) ** 2)
        assert_array_less(obs_err, err)

    spans = [0.05, 0.2, 0.5]
    errs = [0.005, 0.01, 0.1]

    for Model in [MovingAverageSmoother, LinearSmoother]:
        for span, err in zip(spans, errs):
            yield check_model, Model, span, err


def test_sine_cv():
    t, y, dy = make_sine(N=100, err=0.05, rseed=0)
    ytrue = np.sin(np.sort(t)[1:-1])

    def check_model(Model, span, err):
        model = Model(span).fit(t, y, dy)
        yfit = model.cv_values()[1:-1]
        obs_err = np.mean((yfit - ytrue) ** 2)
        assert_array_less(obs_err, err)

    spans = [0.05, 0.2, 0.5]
    errs = [0.005, 0.02, 0.1]

    for Model in [MovingAverageSmoother, LinearSmoother]:
        for span, err in zip(spans, errs):
            yield check_model, Model, span, err


def test_line_linear():
    t, y, dy = make_linear(err=1E-6)
    tfit = np.linspace(0, 10, 40)

    def check_model(span):
        model = LinearSmoother(span).fit(t, y, dy)
        yfit = model.predict(tfit)
        assert_allclose(tfit, yfit, atol=1E-5)

    for span in [0.05, 0.2, 0.5]:
        yield check_model, span



from .. import utils

import numpy as np
from numpy.testing import assert_allclose, assert_


def test_iterable():
    x = [1, 2, 3]
    y = 1
    assert_(utils.iterable(x))
    assert_(not utils.iterable(y))


def test_validate_inputs_nosort(N=10, rseed=0):
    """Test input validation without sorting"""
    rng = np.random.RandomState(rseed)
    t, y = rng.rand(2, N)
    dy = 1

    t_, y_, dy_ = utils.validate_inputs(t, y, dy)
    assert_allclose(t, t_)
    assert_allclose(y, y_)
    assert_allclose(dy, dy_)


def test_validate_inputs_sort(N=10, rseed=0):
    """Test input validation with sorting"""
    rng = np.random.RandomState(rseed)
    t, y = rng.rand(2, N)
    dy = 1

    t_, y_, dy_ = utils.validate_inputs(t, y, dy, sort_by=t)
    isort = np.argsort(t)
    assert_allclose(t[isort], t_)
    assert_allclose(y[isort], y_)
    assert_allclose(dy, dy_)


def test_windowed_sum_fixed(N=10, span=5, rseed=0):
    """Test the windowed sum for a fixed-span"""
    rng = np.random.RandomState(rseed)
    data = rng.rand(3, N)

    for subtract_mid in [True, False]:
        assert_allclose(utils.windowed_sum(*data, span=span,
                                           subtract_mid=subtract_mid),
                        utils.windowed_sum(*data, span=span, slow=True,
                                           subtract_mid=subtract_mid))


def test_windowed_sum_variable(N=10, rseed=0):
    """Test the windowed sum for a variable span"""
    rng = np.random.RandomState(rseed)
    span = rng.randint(3, 6, N)
    data = np.random.random((3, N))

    for subtract_mid in [True, False]:
        assert_allclose(utils.windowed_sum(*data, span=span,
                                           subtract_mid=subtract_mid),
                        utils.windowed_sum(*data, span=span, slow=True,
                                           subtract_mid=subtract_mid))


def make_linear(N=100, err=1E-6, rseed=None):
    rng = np.random.RandomState(rseed)
    t = 10 * rng.rand(N)
    y = t + err * rng.randn(N)
    return t, y, err


def test_constant_data(N=100, rseed=0):
    """Test linear & moving average smoothers on constant data"""
    rng = np.random.RandomState(rseed)
    t = 10 * rng.rand(N)
    y = 1.0
    dy = 1.0
    span = 5

    for method in [utils.moving_average_smooth,
                   utils.linear_smooth]:
        yfit = method(t, y, dy, span=5)
        assert_allclose(y, yfit)


def test_equal_spaced_linear_data(N=100, rseed=0):
    """Test linear & moving average smoothers on equally-spaced linear data"""
    t = np.linspace(0, 10, N)
    y = 1 + 2 * t
    dy = 1
    span = 5

    for method in [utils.moving_average_smooth,
                   utils.linear_smooth]:
        yfit = method(t, y, dy, span=5)
        assert_allclose(y[3:-3], yfit[3:-3])


def test_random_linear_data(N=100, rseed=0):
    """Test linear smoother on linear data"""
    rng = np.random.RandomState(rseed)
    t = 10 * rng.rand(N)
    y = 1 + 2 * t
    dy = 1.0
    span = 5

    yfit = utils.linear_smooth(t, y, dy, span=5)
    assert_allclose(y, yfit)



import numpy as np
from numpy.testing import assert_allclose, assert_array_less, assert_equal

from .. import SuperSmoother


def make_sine(N=100, err=0.05, rseed=None):
    rng = np.random.RandomState(rseed)
    t = 2 * np.pi * rng.rand(N)
    y = np.sin(t) + err * rng.randn(N)
    return t, y, err


def make_linear(N=100, err=1E-6, rseed=None):
    rng = np.random.RandomState(rseed)
    t = 10 * rng.rand(N)
    y = t + err * rng.randn(N)
    return t, y, err


def test_sine():
    t, y, dy = make_sine(N=100, err=0.05, rseed=0)

    tfit = np.linspace(1, 5.3, 50)
    ytrue = np.sin(tfit)

    model = SuperSmoother().fit(t, y, dy)
    yfit = model.predict(tfit)
    obs_err = np.mean((yfit - ytrue) ** 2)
    assert_array_less(obs_err, 0.001)


def test_sine_cv():
    t, y, dy = make_sine(N=100, err=0.05, rseed=0)
    ytrue = np.sin(np.sort(t)[1:-1])

    model = SuperSmoother().fit(t, y, dy)
    yfit = model.cv_values()[1:-1]
    obs_err = np.mean((yfit - ytrue) ** 2)
    assert_array_less(obs_err, 0.001)


def test_line_linear():
    t, y, dy = make_linear(err=1E-6, rseed=0)
    tfit = np.linspace(0, 10, 40)

    model = SuperSmoother().fit(t, y, dy)
    yfit = model.predict(tfit)
    assert_allclose(tfit, yfit, atol=1E-5)



import numpy as np
from .smoother import LinearSmoother
from .utils import linear_smooth


class SuperSmoother(LinearSmoother):
    """
    SuperSmoother, as described in [1]

    [1] Friedman 1984, "A Variable Span Smoother"
        http://www.slac.stanford.edu/cgi-wrap/getdoc/slac-pub-3477.pdf

    There are 3 smoothing levels:
    - tweeter (span=0.05)
    - midrange (span=0.2)
    - woofer (span=0.5)

    9 passes over the data:

    1. Primary smooths for 3 spans & compute residuals
    2. Smooth these residuals with the midrange span
    3. Select span yielding best smoothed residual for each point
    4. Smooth best span estimates with midrange span
    5. Final model: use smoothed span estimates at each point
    """
    default_spans = [0.05, 0.2, 0.5]

    def __init__(self, primary_spans=None):
        if primary_spans is None:
            primary_spans = self.default_spans
        self.primary_spans = np.sort(primary_spans)
        self.middle_span = self.primary_spans[len(primary_spans) // 2]
        self.primary_smooths = [LinearSmoother(span)
                                for span in primary_spans]
        self.smoothed_spans = LinearSmoother(self.middle_span)

    def _fit(self, t, y, dy):
        # 1. Get residuals for each of the primary smooths
        [smoother.fit(t, y, dy, presorted=True)
         for smoother in self.primary_smooths]
        resids = [smoother.cv_residuals()
                  for smoother in self.primary_smooths]

        # 2. Smooth each set of residuals with the midrange
        smoothed_resids = np.array([self.smoothed_spans
                                    .fit(t, abs(resid), 1, False)
                                    .cv_values(cv=False)
                                    for resid in resids])

        # 3. Select span yielding best residual at each point
        best_spans = self.primary_spans[np.argmin(smoothed_resids, 0)]

        # 4. Smooth best span estimates with midrange span
        smoothed_spans = (self.smoothed_spans
                          .fit(t, best_spans, 1, False)
                          .cv_values(cv=False))

        # 5. Use these smoothed span estimates at each point
        self.span = self.smoothed_spans.predict

        # keep these around for efficiencey of evaluating the cv
        self.smoothed_span_vals = smoothed_spans
        LinearSmoother._fit(self, t, y, dy)

    def _cv_values(self, cv=True):
        spans = np.clip(self.smoothed_span_vals * len(self.t), 3, None)
        return linear_smooth(self.t, self.y, self.dy,
                             spans, cv=cv)
